/* Copyright (C) 2006 Anton Fedorov <datacompboy@call2ru.com>

   nss_vserver: provide auth on host according to passwd & shadow files
		in vservers.
   You can find info about vserver at http://linux-vserver.org/
   See README for details
   
   Feel free to use code as you want
 */

#include <pwd.h>
#include <string.h>
#include "config.h"

#define NSS_DB pw
#define NSS_FILES "passwd"
#define NSS_RES passwd

/* Parse line into result, hold strings at buffer.
   Encode usernames with several variants, selected by scantry
   Return values:
	 2: Line parsed, call more with other variants
	 1: Line parsed, no more variants
	 0: Line correct, but no space in buffer
	-1: Line is incorrect
 */
static int parse_line(struct NSS_RES * result,
			   char * buffer, size_t buflen,
			   char * line, size_t linelen,
			   char * vserver, int vserverlen,
			   int scantry)
{
  int colons = 0, i;
  char *err, *colon[6];
  for (i=0; i<linelen; i++) {
	if (line[i]==':') {
	  if (6==colons) return -1;
	  colon[colons++] = &line[i];
	}
  }
  if (colons<6) return -1;
//	result->pw_uid = strtol(colon[2], &err, 10);
//	if (*err != ':') return -1;
//	result->pw_gid = strtol(colon[3], &err, 10);
//	if (*err != ':') return -1;
  {
	int retval = 2;
	int i;
	int bufrest = buflen;
	char * bpos = buffer;
	#define APPEND(a,b) {if(b>=bufrest)return 0;memcpy(bpos,a,b);bpos+=b;bufrest-=b;}
	#define APPENDC(c)  {if(!bufrest--)return 0;*bpos=c; bpos++;}

	result->pw_name = bpos;
	  switch(scantry) {
		case 0: // uname@vserver
		  APPEND(line, colon[0]-line);
		  APPENDC('@');
		  APPEND(vserver, vserverlen);
		  break;
		default: // vserver-uname
		  APPEND(vserver, vserverlen);
		  APPENDC('-');
		  APPEND(line, colon[0]-line);
		  retval = 1;
	  }
	  APPENDC(0);
	result->pw_passwd = bpos;
	  APPEND(colon[0]+1, colon[1]-colon[0]-1);
	  APPENDC(0);
	result->pw_uid = VSUID;
	result->pw_gid = VSGID;
	result->pw_gecos = bpos; /* gecos in form: uname@vserver always */
	  APPEND(line, colon[0]-line);
	  APPENDC('@');
	  APPEND(vserver, vserverlen);
	  APPENDC(0);
	result->pw_dir = bpos;
	  APPEND(ETCVSERVER, strlenETCVSERVER);
	  APPEND(vserver, vserverlen);
	  APPEND("/vdir/.", sizeof("/vdir/.")-1);
	  if (colon[4][1]!='/')
	    APPENDC('/');
	  APPEND(colon[4]+1, colon[5]-colon[4]-1);
	  APPENDC(0);
	result->pw_shell = bpos;
	  APPEND(VSLOGIN, strlenVSLOGIN+1); /* With zero */
	return retval;

	#undef APPEND
	#undef APPENDC
  }
}

/* Select vserver from username */
/* -1 if not vserver linkname */
static int parse_user(char * name,
			   char ** vserver, int * vserverlen,
			   char ** uname, int * unamelen
			  )
{
  int i;
  for (i=0; name[i]; i++) {
	if (name[i]=='@') {
	  int j = strlen(name);
	  *vserver	  = &name[i+1];
	  *vserverlen = j-i-1;
	  *uname	  = name;
	  *unamelen   = i;
	  return 0;
	}
	if (name[i]=='-') {
	  int j = strlen(name);
	  *vserver	  = name;
	  *vserverlen = i;
	  *uname	  = &name[i+1];
	  *unamelen   = j-i-1;
	  return 0;
	}
  }
  return -1;
}

/* Parse line into result, with search of username, that already at result.
   Return values:
	 1: Line parsed fine
	 0: Line fine, but no space available
	-1: Line is incorrect / not one searched
 */
static int parse_line_cmp(struct NSS_RES * result,
				   char * buffer, size_t buflen,
				   char * line, size_t linelen,
				   char * vserver, int vserverlen,
				   char * uname, int unamelen)
{
  int colons = 0, i;
  char *err, *colon[6];
  for (i=0; i<linelen; i++) {
	if (line[i]==':') {
	  if (6==colons) return -1;
	  colon[colons++] = &line[i];
	}
  }
  if (colons<6) return -1;
  if (colon[0]-line != unamelen) return -1;
  if (memcmp(line, uname, unamelen)) return -1;

  {
	char * bpos = buffer + strlen(buffer) + 1;
	int bufrest = buflen - (bpos-buffer);
	#define APPEND(a,b) {if(b>=bufrest)return 0;memcpy(bpos,a,b);bpos+=b;bufrest-=b;}
	#define APPENDC(c)	{if(!bufrest--)return 0;*bpos=c; bpos++;}

	result->pw_name = buffer;
	result->pw_passwd = bpos;
	  APPEND(colon[0]+1, colon[1]-colon[0]-1);
	  APPENDC(0);
	result->pw_uid = VSUID;
	result->pw_gid = VSGID;
	result->pw_gecos = bpos; /* gecos in form: uname@vserver always */
	  APPEND(line, colon[0]-line);
	  APPENDC('@');
	  APPEND(vserver, vserverlen);
	  APPENDC(0);
	result->pw_dir = bpos;
	  APPEND(ETCVSERVER, strlenETCVSERVER);
	  APPEND(vserver, vserverlen);
	  APPEND("/vdir/.", sizeof("/vdir/.")-1);
	  if (colon[4][1]!='/')
	    APPENDC('/');
	  APPEND(colon[4]+1, colon[5]-colon[4]-1);
	  APPENDC(0);
	result->pw_shell = bpos;
	  APPEND(VSLOGIN, strlenVSLOGIN+1); /* With zero */
	  APPENDC(0);
	return 1;

	#undef APPEND
	#undef APPENDC
  }
}

#include "shared.h"

/* Passwd also could search by uid, but that is not our case -- */
/* we provide only dummy UID, so return dummy line */

NSS_STATUS CONCAT3(_nss_vserver_get,NSS_DB,uid_r) (uid_t uid,
									struct NSS_RES *result,
									char *buffer, size_t buflen, int *errnop)
{
  if (uid!=VSUID) {
	*errnop=ENOENT;
	return NSS_STATUS_NOTFOUND;
  } else {
	char * bpos = buffer + strlen(buffer) + 1;
	int bufrest = buflen - (bpos-buffer);
	#define APPEND(a,b) {if(b>=bufrest)return 0;memcpy(bpos,a,b);bpos+=b;bufrest-=b;}
	#define APPENDC(c)	{if(!bufrest--)return 0;*bpos=c; bpos++;}

	result->pw_name = bpos;
	  APPEND("vserver", 8); /* zero incl! */
	result->pw_passwd = bpos;
	  APPEND("x", 2);		/* zero incl! */
	result->pw_uid = VSUID;
	result->pw_gid = VSGID;
	result->pw_gecos = bpos;
	  APPEND("vserver", 8); /* zero incl! */
	result->pw_dir = bpos;
	  APPEND(NONEXISTS, strlenNONEXISTS+1); /* zero incl! */
	result->pw_shell = bpos;
	  APPEND(BINFALSE, strlenBINFALSE+1); /* zero incl! */
	return 1;

	#undef APPEND
	#undef APPENDC
  }

}
